@@ -1,4 +1,5 @@
-/*
+/* -*-mode:c;coding:utf-8; c-basic-offset:2;fill-column:70;c-file-style:"gnu"-*-
+ *
  * Copyright (C) 2009 Arnaud "arnau" Fontaine <arnau@mini-dweeb.org>
  *
  * This  program is  free  software: you  can  redistribute it  and/or
@@ -32,6 +33,24 @@
 #include "plugin.h"
 #include "util.h"
 
+/** Global alpha Pictures cache. This avoids creating an alpha Picture
+    for each window */
+typedef struct __render_alpha_picture_t
+{
+  /** Alpha Picture of the Window */
+  xcb_render_picture_t picture;
+  /** Number of windows currently using this Picture (free the Picture
+      once it reaches 0) */
+  unsigned int reference_counter;
+  /** Alpha Picture opacity */
+  uint16_t opacity;
+  /** Double linked list, perhaps there is a more efficient way, such
+      as hash table but as there are usually not so many opacity
+      pictures... */
+  struct __render_alpha_picture_t *next;
+  struct __render_alpha_picture_t *previous;
+} _render_alpha_picture_t;
+
 /** Information related to Render */
 typedef struct
 {
@@ -45,8 +64,17 @@ typedef struct
   xcb_render_picture_t background_picture;
   /** All Picture formats supported by the screen */
   xcb_render_query_pict_formats_reply_t *pict_formats;
+  /** A8 PictFormat used mainly for alpha Picture (opacity) */
+  xcb_render_pictformat_t a8_pictformat_id;
+  /** ARGB PictFormat used for ARGB Windows */
+  xcb_render_pictformat_t argb_pictformat_id;
   /** Picture Visual supported by the screen */
   xcb_render_pictvisual_t *pictvisual;
+  /** Only the opacity plugins needs such hook ATM, but well something
+      more generic will be written if needed */
+  plugin_t *opacity_plugin;
+  /** Alpha pictures list */
+  _render_alpha_picture_t *alpha_pictures;
 } _render_conf_t;
 
 static _render_conf_t _render_conf;
@@ -56,10 +84,10 @@ typedef struct
 {
   /** Picture associated with the Window Pixmap */
   xcb_render_picture_t picture;
-  /** Alpha Picture of the Window */ 
-  xcb_render_picture_t alpha_picture;
-  /** Alpha Picture opacity of the Window */
-  uint32_t opacity;
+  /** ARGB Window */
+  bool is_argb;
+  /** Pointer to global alpha picture */
+  _render_alpha_picture_t *alpha_picture;
 } _render_window_t;
 
 /** Request label of Render extension for X error reporting, which are
@@ -157,6 +185,9 @@ render_init(void)
   /* Send requests to get the root window background pixmap */ 
   window_get_root_background_pixmap();
 
+  _render_conf.opacity_plugin = plugin_search_by_name("opacity");
+  _render_conf.alpha_pictures = NULL;
+
   return true;
 }
 
@@ -229,12 +260,39 @@ _render_init_root_background(void)
   _render_conf.background_picture = xcb_generate_id(globalconf.connection);
   const uint32_t root_buffer_val = true;
 
-  /* Create a new picture holding the background pixmap */
-  xcb_render_create_picture(globalconf.connection,
-			    _render_conf.background_picture,
-			    root_background_pixmap,
-			    _render_conf.pictvisual->format,
-			    XCB_RENDER_CP_REPEAT, &root_buffer_val);
+  /* Create  a new  picture holding  the background  pixmap through  a
+     'checked' request as  it may fail (for example  when 'display' is
+     used  to set  the background)  and  during startup,  it would  be
+     fatal */
+  xcb_void_cookie_t picture_cookie =
+    xcb_render_create_picture_checked(globalconf.connection,
+                                      _render_conf.background_picture,
+                                      root_background_pixmap,
+                                      _render_conf.pictvisual->format,
+                                      XCB_RENDER_CP_REPEAT, &root_buffer_val);
+
+  /* Check synchronously if the Picture  could be created, if not just
+     set a default background color */
+  xcb_generic_error_t *picture_error = xcb_request_check(globalconf.connection,
+                                                         picture_cookie);
+
+  if(picture_error)
+    {
+      warn("Could not create background Picture, setting a default background "
+           "color (try using another program to set the background?)");
+
+      free(picture_error);
+
+      root_background_pixmap = window_new_root_background_pixmap();
+      root_background_fill = true;
+
+      /* Do not perform any check as it should always succeed */
+      xcb_render_create_picture(globalconf.connection,
+                                _render_conf.background_picture,
+                                root_background_pixmap,
+                                _render_conf.pictvisual->format,
+                                XCB_RENDER_CP_REPEAT, &root_buffer_val);
+    }
 
   if(root_background_fill)
     {
@@ -272,6 +330,16 @@ _render_init_root_picture(void)
       return false;
     }
 
+  /* Used to be computed at each creation of the Window alpha Picture,
+     but seems to be rather costly (as per callgrind) */
+  _render_conf.a8_pictformat_id =
+    xcb_render_util_find_standard_format(_render_conf.pict_formats,
+                                         XCB_PICT_STANDARD_A_8)->id;
+
+  _render_conf.argb_pictformat_id =
+    xcb_render_util_find_standard_format(_render_conf.pict_formats,
+                                         XCB_PICT_STANDARD_ARGB_32)->id;
+
   /* Create Picture associated with the root window */
   {
     _render_conf.picture = xcb_generate_id(globalconf.connection);
@@ -345,7 +413,7 @@ render_reset_background(void)
   xcb_render_free_picture(globalconf.connection,
 			  _render_conf.background_picture);
 
-  /* Send requests to get the root window background pixmap */ 
+  /* Send requests to get the root window background pixmap */
   window_get_root_background_pixmap();
 
   _render_init_root_background();
@@ -354,13 +422,29 @@ render_reset_background(void)
 /** Create the alpha Picture associated  with a window by only filling
  *  it with the alpha channel value
  *
- * \param alpha_picture The alpha Picture of the Window
+ * \param render_window Rendering backend window
  * \param opacity The Window opacity
+ * \return The newly created alpha picture data structure
  */
-static void
-_render_create_window_alpha_picture(xcb_render_picture_t *alpha_picture,
-				    const uint16_t opacity)
+static _render_alpha_picture_t *
+_render_create_window_alpha_picture(_render_window_t *render_window,
+                                    const uint16_t opacity)
 {
+  /* Create a new global alpha_picture */
+  _render_alpha_picture_t *alpha_picture = malloc(sizeof(_render_alpha_picture_t));
+  render_window->alpha_picture = alpha_picture;
+
+  alpha_picture->reference_counter = 1;
+  alpha_picture->opacity = opacity;
+
+  /* Insert it at the beginning of global alpha pictures list */
+  alpha_picture->previous = NULL;
+  alpha_picture->next = _render_conf.alpha_pictures;
+  if(_render_conf.alpha_pictures)
+    _render_conf.alpha_pictures->previous = alpha_picture;
+
+  _render_conf.alpha_pictures = alpha_picture;
+
   const xcb_pixmap_t pixmap = xcb_generate_id(globalconf.connection);
 
   xcb_create_pixmap(globalconf.connection, 8, pixmap,
@@ -368,13 +452,12 @@ _render_create_window_alpha_picture(xcb_render_picture_t *alpha_picture,
 
   const uint32_t create_picture_val = true;
 
-  *alpha_picture = xcb_generate_id(globalconf.connection);
+  alpha_picture->picture = xcb_generate_id(globalconf.connection);
 
   xcb_render_create_picture(globalconf.connection,
-			    *alpha_picture,
+                            alpha_picture->picture,
 			    pixmap,
-			    xcb_render_util_find_standard_format(_render_conf.pict_formats,
-								 XCB_PICT_STANDARD_A_8)->id,
+			    _render_conf.a8_pictformat_id,
 			    XCB_RENDER_CP_REPEAT,
 			    &create_picture_val);
 
@@ -387,10 +470,88 @@ _render_create_window_alpha_picture(xcb_render_picture_t *alpha_picture,
 
   xcb_render_fill_rectangles(globalconf.connection,
 			     XCB_RENDER_PICT_OP_SRC,
-			     *alpha_picture,
+                             alpha_picture->picture,
 			     color, 1, &rect);
 
   xcb_free_pixmap(globalconf.connection, pixmap);
+
+  return alpha_picture;
+}
+
+/** Decrement  the reference  counter of  the alpha  picture currently
+ *  associated  with  the  given  rendering  backend  window.  If  the
+ *  reference counter reaches  0, then free its  Picture and allocated
+ *  memory, and update the global alpha picture list.
+ *
+ * \param render_window Rendering backend window
+ */
+static void
+_render_unref_window_alpha_picture(_render_window_t *render_window)
+{
+  if(render_window->alpha_picture->reference_counter == 1)
+    {
+      xcb_render_free_picture(globalconf.connection,
+                              render_window->alpha_picture->picture);
+
+      if(render_window->alpha_picture->previous)
+        render_window->alpha_picture->previous->next =
+          render_window->alpha_picture->next;
+
+      if(render_window->alpha_picture->next)
+        render_window->alpha_picture->next->previous =
+          render_window->alpha_picture->previous;
+
+      if(render_window->alpha_picture == _render_conf.alpha_pictures)
+        _render_conf.alpha_pictures = _render_conf.alpha_pictures->next;
+
+      util_free(&render_window->alpha_picture);
+    }
+  else
+    {
+      --render_window->alpha_picture->reference_counter;
+      render_window->alpha_picture = NULL;
+    }
+}
+
+/** Get the alpha Picture associated  with the given rendering backend
+ *  window, and create it if it does not already exist.
+ *
+ * \param render_window Rendering backend window
+ * \param opacity New opacity value
+ * \return Render Picture XID
+ */
+static xcb_render_picture_t
+_render_get_window_alpha_picture(_render_window_t *render_window,
+                                 const uint16_t opacity)
+{
+  /* Return the Picture XID if  the opacity has not changed, otherwise
+     decrement the reference counter of the previous alpha Picture */
+  if(render_window->alpha_picture)
+    {
+      if(render_window->alpha_picture->opacity == opacity)
+        return render_window->alpha_picture->picture;
+      else
+        _render_unref_window_alpha_picture(render_window);
+    }
+
+  /* Opaque Window, do nothing */
+  if(opacity == UINT16_MAX)
+    return XCB_NONE;
+
+  /* The opacity property  value has changed, try to get  one from the
+     cache if possible */
+  _render_alpha_picture_t *alpha_picture;
+  for(alpha_picture = _render_conf.alpha_pictures; alpha_picture;
+      alpha_picture = alpha_picture->next)
+    if(alpha_picture->opacity == opacity)
+      {
+        ++alpha_picture->reference_counter;
+        render_window->alpha_picture = alpha_picture;
+        return alpha_picture->picture;
+      }
+
+  /* Otherwise, create a new one */
+  return _render_create_window_alpha_picture(render_window, opacity)->picture;
 }
 
 /** Paint the root background to the buffer Picture */
@@ -435,6 +596,9 @@ render_paint_window(window_t *window)
 	xcb_render_util_find_visual_format(_render_conf.pict_formats,
 					   window->attributes->visual);
 
+      render_window->is_argb = (window_pictvisual->format ==
+                                _render_conf.argb_pictformat_id);
+
       xcb_render_create_picture(globalconf.connection,
 				render_window->picture, window->pixmap,
 				window_pictvisual->format,
@@ -442,80 +606,54 @@ render_paint_window(window_t *window)
 				&create_picture_val);
     }
 
-  uint8_t render_composite_op;
-  xcb_render_picture_t window_alpha_picture;
+  uint8_t render_composite_op = XCB_RENDER_PICT_OP_SRC;
+  xcb_render_picture_t alpha_picture = XCB_NONE;
 
-  {
-    /* Only  the  opacity  plugins  needs  such  hook  ATM,  but  well
-       something more generic will be written if needed */
-    plugin_t *opacity_plugin = plugin_search_by_name("opacity");
-    uint16_t opacity;
+  if(render_window->is_argb)
+    render_composite_op = XCB_RENDER_PICT_OP_OVER;
+  if(_render_conf.opacity_plugin &&
+          _render_conf.opacity_plugin->vtable->window_get_opacity)
+    {
+      alpha_picture =
+        _render_get_window_alpha_picture(render_window,
+                                         (*_render_conf.opacity_plugin->vtable->window_get_opacity)(window));
 
-    if(opacity_plugin && opacity_plugin->vtable->window_get_opacity &&
-       (opacity = (*opacity_plugin->vtable->window_get_opacity)(window)) != UINT16_MAX)
-      {
-        /* Re-create the alpha Picture if the opacity was changed */
-        if(render_window->alpha_picture &&
-           render_window->opacity != opacity)
-          {
-            xcb_render_free_picture(globalconf.connection,
-                                    render_window->alpha_picture);
+      if(alpha_picture != XCB_NONE)
+        render_composite_op = XCB_RENDER_PICT_OP_OVER;
+    }
 
-            render_window->alpha_picture = XCB_NONE;
-          }
+  /* For  non-rectangular  Windows, clip  the  Window  Picture to  its
+     shaped Region to paint  them properly (otherwise for applications
+     such  as  xeyes,  garbage  pixels are  shown  as  RenderComposite
+     expects a rectangular area)
 
-	if(!render_window->alpha_picture)
-          {
-            _render_create_window_alpha_picture(&render_window->alpha_picture,
-                                                opacity);
+     \todo: Should ShapeNotify be handled as well?
+  */
+  if(!window_is_rectangular(window))
+    {
+      xcb_xfixes_region_t shape_region = window_get_region(window, false, false);
 
-            render_window->opacity = opacity;
-          }
+      xcb_xfixes_set_picture_clip_region(globalconf.connection,
+                                         render_window->picture,
+                                         shape_region,
+                                         window->geometry->border_width,
+                                         window->geometry->border_width);
 
-	render_composite_op = XCB_RENDER_PICT_OP_OVER;
-	window_alpha_picture = render_window->alpha_picture;
-      }
-    else
-      {
-        /* Make  sure the  alpha  Picture is  freed  after the  Window
-           turned opaque but was previously translucent */
-        if(render_window->alpha_picture != XCB_NONE)
-          xcb_render_free_picture(globalconf.connection,
-                                  render_window->alpha_picture);
-
-	render_composite_op = XCB_RENDER_PICT_OP_SRC;
-	window_alpha_picture = XCB_NONE;
-      }
-  }
-
-  /* Only paint  from the  Window Region, otherwise  it does  not work
-     properly for non-rectangular windows such as xeyes */
-  xcb_xfixes_set_picture_clip_region(globalconf.connection,
-                                     _render_conf.buffer_picture,
-                                     window->region, 0, 0);
-
-  debug("%x: Clipping to %dx%d +%d+%d", window->id,
-        (uint16_t) (window->geometry->width + window->geometry->border_width * 2),
-        (uint16_t) (window->geometry->height + window->geometry->border_width * 2),
-        window->geometry->x, window->geometry->y);
+      xcb_xfixes_destroy_region(globalconf.connection, shape_region);
+    }
 
   xcb_render_composite(globalconf.connection,
 		       render_composite_op,
-		       render_window->picture, window_alpha_picture, _render_conf.buffer_picture,
+		       render_window->picture,
+                       alpha_picture,
+                       _render_conf.buffer_picture,
 		       0, 0, 0, 0,
 		       window->geometry->x,
 		       window->geometry->y,
-		       (uint16_t) (window->geometry->width + window->geometry->border_width * 2),
-		       (uint16_t) (window->geometry->height + window->geometry->border_width * 2));
-
-  /* In software rendering, there is no need to reset the clipping
-     region but with GPU accelerated rendering, it's completely buggy
-     (or is it an expected behavior?)  with most drivers (tested with
-     Intel, ATI and Nvidia GPUs). For example with Intel driver, only
-     the first window is painted... */
-  xcb_xfixes_set_picture_clip_region(globalconf.connection,
-                                     _render_conf.buffer_picture,
-                                     XCB_NONE, 0, 0);
+		       (uint16_t) (window->geometry->width +
+                                   window->geometry->border_width * 2),
+		       (uint16_t) (window->geometry->height +
+                                   window->geometry->border_width * 2));
 }
 
 /** Routine to  paint everything on  the root Picture, it  just paints
@@ -600,7 +738,12 @@ render_free_window_pixmap(window_t *window)
 static void
 render_free_window(window_t *window)
 {
-  free((_render_window_t *) window->rendering);
+  _render_window_t *render_window = (_render_window_t *) window->rendering;
+
+  if(render_window && render_window->alpha_picture)
+    _render_unref_window_alpha_picture(render_window);
+
+  free(render_window);
 }
 
 /** Called on dlclose()  and free all the resources  allocated by this
